/*
    XMMSd - A web control plug-in for XMMS
    Copyright (C) 2002  Fredrik Rambris <boost@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* server.c - This is the core part. Herein are all functions for parsing
 * requests, listening for messages etc.
 */

#include <config.h>
#include <stdio.h>
#include <glib.h>
#include <sys/time.h>

#include "types.h"
#include "util.h"
#include "base64.h"
#include "taglist.h"
#include "server.h"

/* The different phases of the communication */

/* Read the first line and parse the request. */
#define PHASE_REQUEST 0

/* Collect all headers until a blank line is received */
#define PHASE_HEADERS 1

/* If there are any data after the headers (such is the case in a POST request
 * this data is read in */
#define PHASE_BODY 2

/* Build and send the response */
#define PHASE_RESPONSE 3

/* If no response handler is set this is assigned to the task */
void boche_standard_response( struct Request *req )
{
	g_string_sprintfa( req->response,
			"<HTML>\n"
			"<HEAD>\n"
			"<TITLE>Boche web server</TITLE>\n"
			"</HEAD>\n"
			"<BODY>\n"
			"<H1>Boche web server</H1>\n"
			"You've reached the Boche web server and the programmer haven't assigned a response handler of her own\n"
			"<HR>\n"
			"Boche %s by <A HREF=\"http://fredrik.rambris.com/\">Fredrik Rambris</A>\n"
			"</BODY>\n"
			"</HTML>\n",
			BOCHE_VERSION
	);
}

gboolean free_node( gpointer key, gpointer value, gpointer user_data );
gint ngets( gchar *buf, gint len, gint socket );
gboolean nprintf( gint socket, gchar *fmt, ... );
gint nread( gchar *buf, gint len, gint socket );

struct Server *server_newA( struct TagItem *tags );
struct Server *server_new( Tag tag1, ... );
void cancel_request( gpointer data, gpointer user_data );
void server_cleanup( void *arg );
void server_free( struct Server *serv );
gboolean server_findfreeslot( struct Request *req );
void server_releaseslot( struct Request *req );
void *server( void *arg );
struct Request *request_new( struct Server *serv, gint socket );
void request_free( struct Request *req );
void *request( void *arg );

/* Used to free all nodes in hash table */
gboolean free_node( gpointer key, gpointer value, gpointer user_data )
{
	gboolean is_gstring=GPOINTER_TO_INT( user_data );
	GString *gs=value;
	g_free( key );
	if( is_gstring )
	{
		g_free( gs->str );
	}
	g_free( value );
	return TRUE;
}

/* fprintf replacement. Sends to socket instead. n=network */
gboolean nprintf( gint socket, gchar *fmt, ... )
{
	va_list ap;
	gchar *str;
	int ret;
	size_t len;
	gboolean result=FALSE;
	if( socket<0 || !fmt ) pthread_exit( NULL );
	va_start(ap, fmt);
	if( ( str=g_strdup_vprintf( fmt, ap ) ) )
	{
		len=strlen( str );
		ret=send( socket, str, len, MSG_NOSIGNAL );
		if( ret==len ) result=TRUE;
		g_free( str );
	}
	va_end(ap);
	if( !result ) pthread_exit( NULL ); /* If we get an error just kill the thread */
	return result;
}

/* fgets replacement. */
gint ngets( gchar *buf, gint len, gint socket )
{
	fd_set rfds;
	struct timeval tv;
	int retval;
	int pos=0;
	if( !buf ) return 0;
	memset( buf, 0, len ); /* Clear the string */
	while( TRUE )
	{
		if( pos>=(len-1) ) return pos;
		FD_ZERO(&rfds);
		FD_SET(socket, &rfds);
		
		tv.tv_sec = 30;
		tv.tv_usec = 0;
		retval = select(socket+1, &rfds, NULL, NULL, &tv);
		if( !retval ) break;
		if( FD_ISSET( socket, &rfds ) )
		{
			retval=recv( socket, buf+pos, 1, MSG_NOSIGNAL );
			if( retval!=1 ) break;
			if( buf[pos]=='\n' )
			{
				return pos+1;
			}
			pos+=retval;
		}
	}
	pthread_exit( NULL ); /* Just drop out */
}

/* fread replacement. We skip the nmemb. */
gint nread( gchar *buf, gint len, gint socket )
{
	fd_set rfds;
	struct timeval tv;
	int retval;
	int pos=0;
	if( !buf ) return 0;
	memset( buf, 0, len ); /* Clear the string */
	while( TRUE )
	{
		if( pos==len ) return pos; /* Done! */
		FD_ZERO(&rfds);
		FD_SET(socket, &rfds);
		
		tv.tv_sec = 30;
		tv.tv_usec = 0;
		retval = select(socket+1, &rfds, NULL, NULL, &tv);
		if( !retval ) break; /* Timeout */
		if( FD_ISSET( socket, &rfds ) )
		{
			retval=recv( socket, buf+pos, 1, MSG_NOSIGNAL ); /* We read one char at a time */
			if( retval!=1 ) break;
			pos+=retval;
		}
	}
	pthread_exit( NULL ); /* Just drop out */
}

/* Set up and start a new server thread */
struct Server *server_newA( struct TagItem *tags )
{
	gint port=GetTagData( SERVER_PORT, 8335, tags );
	gchar *bind_to=GINT_TO_POINTER( GetTagData( SERVER_BIND, GPOINTER_TO_INT( "127.0.0.1" ), tags ) );
	gchar *server_hostname=GINT_TO_POINTER( GetTagData( SERVER_HOSTNAME, 0, tags ) );
	void (*response_handler)(struct Request *)=GINT_TO_POINTER( GetTagData( SERVER_RESPONSE, GPOINTER_TO_INT( boche_standard_response ), tags ) );
	gboolean (*authentication_handler)(struct Request *)=GINT_TO_POINTER( GetTagData( SERVER_AUTHENTICATION, 0, tags ) );
	gpointer user_data=GINT_TO_POINTER( GetTagData( SERVER_USERDATA, 0, tags ) );
	gint numclients=GetTagData( SERVER_NUMCLIENTS, 10, tags );
	Tag ourTags[]={ SERVER_PORT, SERVER_BIND, SERVER_RESPONSE, SERVER_AUTHENTICATION, SERVER_HOSTNAME, SERVER_USERDATA };
	struct Server serv, *servptr;
	static gint yes=1;
	memset( &serv, 0, sizeof( struct Server ) );
	serv.port=port;
	serv.numclients=numclients;
	if( ( serv.servsock=socket( AF_INET, SOCK_STREAM, 0 ) ) == -1 )
	{
		perror( "socket" );
		return NULL;
	}

	if( setsockopt( serv.servsock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int) ) == -1)
	{
		perror("setsockopt");
		close( serv.servsock );
		return NULL;
    }

	serv.myaddr.sin_family = AF_INET;
	if( strcmp(bind_to,"*") )
	{
		serv.address = inet_addr( bind_to );
		if( serv.address == INADDR_NONE )
		{
			fprintf( stderr, "Invalid address supplied\n" );
			close( serv.servsock );
			return NULL;
		}
		serv.myaddr.sin_addr.s_addr=serv.address;
	}
	else
	{
		serv.myaddr.sin_addr.s_addr=INADDR_ANY;
	}

	serv.myaddr.sin_port=htons(port);
	memset( &(serv.myaddr.sin_zero), '\0', 8 );
	if( bind( serv.servsock, (struct sockaddr *)&(serv.myaddr), sizeof( serv.myaddr ) ) == -1 )
	{
		perror( "bind" );
		close( serv.servsock );
		return NULL;
	}

	/* We use the same default as XMMS */
	if( listen( serv.servsock, 100 ) == -1 )
	{
		perror( "listen" );
		close( serv.servsock );
		return NULL;
	}

	if( !server_hostname )
	{
		server_hostname=g_new0( gchar, 1024 );
		if( gethostname( server_hostname, 1024 ) )
		{
			g_free( server_hostname );
			server_hostname=NULL;
			close( serv.servsock );
			return NULL;
		}
		serv.server_hostname=g_strdup( server_hostname ); /* We don't want to waste 1024 bytes on a string that is shorter */
		g_free( server_hostname );
	}
	else
	{
		/* If you supplied a string in the taglist be sure to free it if it was
		 * dynamicly allocated */
		serv.server_hostname=g_strdup( server_hostname );
	}

	if( ( servptr=g_new0( struct Server, 1 ) ) )
	{
		pthread_mutex_t mut=PTHREAD_MUTEX_INITIALIZER;
		memcpy( servptr, &serv, sizeof( struct Server ) );
		memcpy( &(servptr->clients_mutex), &mut, sizeof( pthread_mutex_t ) );
		servptr->clients=g_new0( struct Request *, servptr->numclients );
		servptr->response_handler=response_handler;
		servptr->authentication_handler=authentication_handler;
		FilterTagItems( tags, ourTags, TAGFILTER_NOT );
		servptr->taglist=CloneTagItems( tags );
		servptr->user_data=user_data;
		servptr->running=TRUE;
		if( !pthread_create( &(servptr->thread), NULL, server, (void *)servptr ) ) return servptr;
		g_free( servptr );
		close( serv.servsock );
	}
	return NULL;
}

/* Varargs stub */
struct Server *server_new( Tag tag1, ... )
{
	return server_newA( (struct TagItem *)&tag1 );
}

void cancel_request( gpointer data, gpointer user_data )
{
	struct Request *req=( struct Request *)data;
	if( req ) pthread_cancel( req->thread );
}

void server_cleanup( void *arg )
{
	struct Server *serv=( struct Server *)arg;
	gint slotno;
	if( !serv ) return;
	serv->running=FALSE;
	if( serv->servsock!=-1 )
	{
		close( serv->servsock );
		serv->servsock=-1;
	}
	for( slotno=0 ; slotno<serv->numclients ; slotno++ )
	{
		if( serv->clients[slotno] )
		{
			pthread_cancel( serv->clients[slotno]->thread );
		}
	}
	usleep(100000); /* Allow some time for requests to finish freeing up */
	g_free( serv->clients );
}

void server_free( struct Server *serv )
{
	if( !serv ) return;
	serv->running=FALSE;
	pthread_cancel( serv->thread );
	server_cleanup( (void *)serv );
	g_free( serv );
}

gboolean server_findfreeslot( struct Request *req )
{
	gint ret=FALSE;
	gint slotno;
	if( !req ) return ret;
	if( !req->server ) return ret;
	if( !req->server->numclients ) return ret;
	pthread_mutex_lock( &(req->server->clients_mutex) );
	for( slotno=0 ; slotno<(req->server->numclients) ; slotno++ )
	{
		if( !req->server->clients[slotno] )
		{
			req->server->clients[slotno]=req;
			req->slotno=slotno;
			ret=TRUE;
			break;
		}
	}
	pthread_mutex_unlock( &(req->server->clients_mutex) );
	return ret;
}

void server_releaseslot( struct Request *req )
{
	if( !req ) return;
	if( !req->server ) return;
	if( req->slotno==-1 ) return;
	pthread_mutex_lock( &(req->server->clients_mutex) );
	req->server->clients[req->slotno]=NULL;
	pthread_mutex_unlock( &(req->server->clients_mutex) );
	req->slotno=-1;
}

void *server( void *arg )
{
	struct Server *serv=( struct Server *)arg;
	struct Request *req=NULL;
	gint socket;
	struct sockaddr_in addr;
	int addrlen;
	if( !serv ) return NULL;


	while( serv->running )
	{
		if( ( socket=accept( serv->servsock, (struct sockaddr *)&addr, &addrlen ) )==-1 )
		{
			break;
		}
		if( ( req=request_new( serv, socket ) ) )
		{
			req->clientaddr=g_strdup( inet_ntoa( addr.sin_addr ) );
			if( pthread_create( &(req->thread), NULL, request, (void *)req ) )
			{
				request_free( req );
			}
		}
	}
	server_cleanup( (void *)serv );
}

/* Allocate and initialize a request structure */
struct Request *request_new( struct Server *serv, gint socket )
{
	struct Request *req;
	static gchar werefull[]="HTTP/1.0 500 Internal server error\r\nContent-type: text/html\r\n\r\n<HTML><BODY>The Internet is full. Go away!</BODY></HTML>";
	if( !serv || socket==-1 ) return NULL;
	if( !serv->running )
	{
		/* Won't create a new request if we're in the middle of closing down */
		close( socket );
		return( NULL );
	}
	req=g_new0( struct Request, 1 );
	if( !req ) return req;
	req->server=serv;
	if( !server_findfreeslot( req ) )
	{
		send( socket, werefull, strlen( werefull ), 0 );
		close( socket );
		g_free( req );
		return NULL;
	}
	req->socket=socket;
	req->get_entries=g_hash_table_new( g_str_hash, g_str_equal );
	req->post_entries=g_hash_table_new( g_str_hash, g_str_equal );
	req->headers=g_hash_table_new( g_str_hash, g_str_equal );
	req->response=g_string_sized_new( 500 );
	req->content_type=g_string_new( "text/html; charset=iso-8859-1" );
	req->ret_code=200;
	return req;
}

void request_cleanup( void *arg )
{
	request_free( (struct Request *)arg );
}

/* Free up itself as well as it's resources */
void request_free( struct Request *req )
{
	GList *node;
	if( !req ) return;
	server_releaseslot( req );
	if( req->socket!=-1 ) close( req->socket );
	req->socket=-1;
	if( req->requestline ) g_free( req->requestline );
	req->requestline=NULL;
	if( req->path ) g_free( req->path );
	req->path=NULL;
	if( req->postdata ) g_free( req->postdata );
	req->postlength=0;
	req->method=0;
	if( req->get_entries )
	{
		g_hash_table_foreach_remove( req->get_entries, free_node, GINT_TO_POINTER( 1 ) );
		g_hash_table_destroy( req->get_entries );
		req->get_entries=NULL;
	}
	if( req->post_entries )
	{
		g_hash_table_foreach_remove( req->post_entries, free_node, GINT_TO_POINTER( 1 ) );
		g_hash_table_destroy( req->post_entries );
		req->post_entries=NULL;
	}
	if( req->headers )
	{
		g_hash_table_foreach_remove( req->headers, free_node, NULL );
		g_hash_table_destroy( req->headers );
		req->headers=NULL;
	}
	if( req->auth_user )
	{
		g_free( req->auth_user );
	}
	req->auth_user=NULL;
	if( req->auth_pass )
	{
		g_free( req->auth_pass );
	}
	req->auth_pass=NULL;

	if( req->response ) g_string_free( req->response, TRUE );
	req->response=NULL;
	for( node=req->response_headers; node; node=node->next )
	{
		if( node->data ) g_free( node->data );
		node->data=NULL;
	}
	g_list_free( req->response_headers );
	req->response_headers=NULL;
	if( req->content_type ) g_string_free( req->content_type, TRUE );
	req->content_type=NULL;
	if( req->clientaddr ) g_free( req->clientaddr );
	req->clientaddr=NULL;
	g_free( req );
}

void send_header( gpointer data, gpointer user_data )
{
	struct Request *req=user_data;
	gchar *hdr=data;
	if( !req || !hdr ) return;
	nprintf( req->socket, "%s\r\n", hdr );
}

void *request( void *arg )
{
	struct Request *req=( struct Request *)arg;
	char buffer[1024];
	int phase;
	int buflen;
	gchar **vptr=NULL;
	gchar **command=NULL;
	char *cmdptr, *tptr;
	int count;

	if( !req ) return NULL;
	if( !req->server ) return NULL;
	if( !req->server->response_handler ) return NULL;
	req->disconnect=FALSE;
	phase=PHASE_REQUEST;

	pthread_cleanup_push( request_cleanup, (void *)req );
	while( !req->disconnect )
	{
		switch( phase )
		{
			case PHASE_REQUEST: /* Collect and parse request */
				if( !ngets( buffer, 1024, req->socket ) )
				{
					req->disconnect=TRUE;
					break;
				}
				if( strchr( buffer, '\n' ) ) stripws( buffer );
				else
				{
					req->disconnect=TRUE;
					break;
				}

				if( !strlen( buffer ) )
				{
					req->disconnect=TRUE;
					break;
				}
				req->requestline=g_strdup( buffer );
				if( ( vptr=g_strsplit( buffer, " ", 3 ) ) )
				{
					for( count=0; vptr[count]; count++ ); /* Get the number of entries */
					if( count>1 )
					{
						if( !strcmp( vptr[0], "GET" ) )
							req->method=METHOD_GET;
						else if( !strcmp( vptr[0], "POST" ) )
							req->method=METHOD_POST;
						/* More method probably could be implemented here */
						else
						{
							req->disconnect=TRUE;
							break;
						}
						if( count>2 ) /* Get the HTTP version used */
						{
							sscanf( vptr[2], "HTTP/%d.%d", &req->http_major, &req->http_minor );
						}
						if( ( cmdptr=strchr( vptr[1], '?' ) ) )
						{
							cmdptr[0]='\0';
							cmdptr++;
							request_querydecode( req->get_entries, cmdptr );
						}
						req->path=g_strdup( vptr[1] );
						phase=PHASE_HEADERS;
					}
					else req->disconnect=TRUE;
					g_strfreev( vptr );
				} else req->disconnect=TRUE;
				break;

			case PHASE_HEADERS: /* Collect headers. */
				if( !ngets( buffer, 1024, req->socket ) )
				{
					req->disconnect=TRUE;
					break;
				}
				if( strchr( buffer, '\n' ) ) stripws( buffer );
				else
				{
					req->disconnect=TRUE;
					break;
				}
				if( !strlen( buffer ) )
				{
					/* If we haven't received a Host: header set our server_hostname to
					 * a duplicate of the one in struct Server. Unless the client uses
					 * HTTP/1.1 or later where we must reply with a 400 Bad Request
					 * acording to RFC-2616. */
					if( !req->server_hostname )
					{
						if( req->http_major>=1 && req->http_minor>=1 )
						{
							phase=PHASE_RESPONSE;
							req->ret_code=400;
							break;
						}
						req->server_hostname=g_strdup( req->server->server_hostname );
					}
					if( !req->server_port ) req->server_port=req->server->port;
/* Darn shame we can't nest comments! */
#ifdef LOG
					cmdptr=(gchar *)g_hash_table_lookup( req->headers, "User-Agent" ); /* Reusing variable */
					g_message( "%s \"%s\" \"%s\"", req->clientaddr, req->requestline, (cmdptr?cmdptr:"-") );
#endif
					/* Switch phasers to stunn */
					if( req->method==METHOD_GET ) phase=PHASE_RESPONSE;
					else if( req->method==METHOD_POST ) phase=PHASE_BODY;
					break;
				}
				if( ( cmdptr=strchr( buffer, ':' ) ) )
				{
					cmdptr[0]='\0';
					cmdptr++;
					stripws( buffer );
					stripws( cmdptr );
					g_hash_table_insert( req->headers, g_strdup( buffer ), g_strdup( cmdptr ) );
					if( !strcasecmp( buffer, "Content-Length" ) ) /* Used in phase 3 (POST) */
					{
						req->postlength=atoi( cmdptr );
					}
					else if( !strcasecmp( buffer, "Authorization" ) )
					{
						if( ( vptr=g_strsplit( cmdptr, " ", 8 ) ) )
						{
							if( !strcasecmp( vptr[0], "Basic" ) )
							{
								base64_decode( vptr[1] );

								if( ( cmdptr=strchr( vptr[1], ':' ) ) )
								{
									cmdptr[0]='\0';
									cmdptr++;
									req->auth_user=g_strdup( vptr[1] );
									req->auth_pass=g_strdup( cmdptr );
								}
							}
							g_strfreev( vptr );
						}
					}
					else if( !strcasecmp( buffer, "Host" ) )
					{
						/* If the client for some reason sends two Host headers */
						if( ( tptr=strchr( cmdptr, ':' ) ) )
						{
							tptr[0]='\0';
							tptr++;
							req->server_port=atoi( tptr );
						}
						else req->server_port=80; /* According to RFC-2616 */
						if( req->server_hostname ) g_free( req->server_hostname );
						req->server_hostname=g_strdup( cmdptr );
					}
				}
				break;

			case PHASE_BODY: /* Get body (from POST requests) */
				if( req->postlength>0 )
				{
					req->postdata=g_new0( char, req->postlength );
					nread( req->postdata, req->postlength, req->socket );
					if( ( cmdptr=(gchar *)g_hash_table_lookup( req->headers, (gpointer)"Content-Type" ) ) )
					{
						if( !strcasecmp( cmdptr, "application/x-www-form-urlencoded" ) )
						{
							request_querydecode( req->post_entries, req->postdata );
						}
					}
					// ADD CHECKING HERE! DO THIS BETTER!
				}
				phase=PHASE_RESPONSE;
				break;

			case PHASE_RESPONSE: /* Handle response */
				if( req->server->authentication_handler )
				{
					req->server->authentication_handler( req );
				}
				/* We don't want to bother the response handler with errors. We'll implement
				 * an error handler later instead */
				if( req->ret_code==200 )
				{
					req->server->response_handler( req );
				}
				if( req->http_major || req->http_minor ) /* Send HTTP/x.x response */
				{
					/* I just grabbed all these from RFC 2616. We don't handle all these correct yet*/
					if( req->ret_code==100 ) nprintf( req->socket, "HTTP/1.0 100 Continue\r\n" );
					else if( req->ret_code==101 ) nprintf( req->socket, "HTTP/1.0 101 Switching Protocols\r\n" );
					else if( req->ret_code==200 ) nprintf( req->socket, "HTTP/1.0 200 OK\r\n" );
					else if( req->ret_code==201 ) nprintf( req->socket, "HTTP/1.0 201 Created\r\n" );
					else if( req->ret_code==202 ) nprintf( req->socket, "HTTP/1.0 202 Accepted\r\n" );
					else if( req->ret_code==203 ) nprintf( req->socket, "HTTP/1.0 203 Non-Authoritative Information\r\n" );
					else if( req->ret_code==204 ) nprintf( req->socket, "HTTP/1.0 204 No Content\r\n" );
					else if( req->ret_code==205 ) nprintf( req->socket, "HTTP/1.0 205 Reset Content\r\n" );
					else if( req->ret_code==206 ) nprintf( req->socket, "HTTP/1.0 206 Partial Content\r\n" );
					else if( req->ret_code==300 ) nprintf( req->socket, "HTTP/1.0 300 Multiple Choices\r\n" );
					else if( req->ret_code==301 ) nprintf( req->socket, "HTTP/1.0 301 Moved Permanently\r\n" );
					else if( req->ret_code==302 ) nprintf( req->socket, "HTTP/1.0 302 Found\r\n" );
					else if( req->ret_code==303 ) nprintf( req->socket, "HTTP/1.0 303 See Other\r\n" );
					else if( req->ret_code==304 ) nprintf( req->socket, "HTTP/1.0 304 Not Modified\r\n" );
					else if( req->ret_code==305 ) nprintf( req->socket, "HTTP/1.0 305 Use Proxy\r\n" );
					else if( req->ret_code==307 ) nprintf( req->socket, "HTTP/1.0 307 Temporary Redirect\r\n" );
					else if( req->ret_code==400 )
					{
						nprintf( req->socket, "HTTP/1.0 400 Bad Request\r\n" );
						g_string_sprintf( req->response,
							"<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n"
							"<HTML><HEAD>\n"
							"<TITLE>400 Bad Request</TITLE>\n"
							"</HEAD><BODY>\n"
							"<H1>Bad Request</H1>\n"
							"Your browser sent a request that this server could not understand.<P>\n"
							"client sent HTTP/1.1 request without hostname (see RFC2616 section 14.23): /<P>\n"
						);
					}
					else if( req->ret_code==401 )
					{
						nprintf( req->socket, "HTTP/1.0 401 Unauthorized\r\n" );
						/* It's a responsibility of the response handlers to add this
						   nprintf( req->socket, "WWW-Authenticate: Basic realm=\"Boche\"\r\n" ); */
					}
					else if( req->ret_code==402 ) nprintf( req->socket, "HTTP/1.0 402 Payment Required\r\n" );
					else if( req->ret_code==403 ) nprintf( req->socket, "HTTP/1.0 403 Forbidden\r\n" );
					else if( req->ret_code==404 )
					{
						nprintf( req->socket, "HTTP/1.0 404 Not Found\r\n" );
						if( !strlen( req->response->str ) ) g_string_sprintf( req->response, "<HTML><HEAD><TITLE>404 not found</TITLE></HEAD><BODY><H1>Not found</H1>The requested path %s doesn't exist on this server</BODY></HTML>", req->path );
					}
					else if( req->ret_code==405 ) nprintf( req->socket, "HTTP/1.0 405 Method Not Allowed\r\n" );
					else if( req->ret_code==406 ) nprintf( req->socket, "HTTP/1.0 406 Not Acceptable\r\n" );
					else if( req->ret_code==407 ) nprintf( req->socket, "HTTP/1.0 407 Proxy Authentication Required\r\n" );
					else if( req->ret_code==408 ) nprintf( req->socket, "HTTP/1.0 408 Request Timeout\r\n" );
					else if( req->ret_code==409 ) nprintf( req->socket, "HTTP/1.0 409 Conflict\r\n" );
					else if( req->ret_code==410 ) nprintf( req->socket, "HTTP/1.0 410 Gone\r\n" );
					else if( req->ret_code==411 ) nprintf( req->socket, "HTTP/1.0 411 Length Required\r\n" );
					else if( req->ret_code==412 ) nprintf( req->socket, "HTTP/1.0 412 Precondition Failed\r\n" );
					else if( req->ret_code==413 ) nprintf( req->socket, "HTTP/1.0 413 Request Entity Too Large\r\n" );
					else if( req->ret_code==414 ) nprintf( req->socket, "HTTP/1.0 414 Request-URI Too Long\r\n" );
					else if( req->ret_code==415 ) nprintf( req->socket, "HTTP/1.0 415 Unsupported Media Type\r\n" );
					else if( req->ret_code==416 ) nprintf( req->socket, "HTTP/1.0 416 Requested Range Not Satisfiable\r\n" );
					else if( req->ret_code==417 ) nprintf( req->socket, "HTTP/1.0 417 Expectation Failed\r\n" );
					else if( req->ret_code==500 ) nprintf( req->socket, "HTTP/1.0 500 Internal Server Error\r\n" );
					else if( req->ret_code==501 ) nprintf( req->socket, "HTTP/1.0 501 Not Implemented\r\n" );
					else if( req->ret_code==502 ) nprintf( req->socket, "HTTP/1.0 502 Bad Gateway\r\n" );
					else if( req->ret_code==503 ) nprintf( req->socket, "HTTP/1.0 503 Service Unavailable\r\n" );
					else if( req->ret_code==504 ) nprintf( req->socket, "HTTP/1.0 504 Gateway Timeout\r\n" );
					else if( req->ret_code==505 ) nprintf( req->socket, "HTTP/1.0 505 HTTP Version Not Supported\r\n" );
					nprintf( req->socket, "Content-Type: %s\r\n", req->content_type->str );
					nprintf( req->socket, "Content-Length: %d\r\n", req->response->len );
					/* I've dubbed the server part (enginge sorta) to Boche */
					nprintf( req->socket, "Server: Boche/" BOCHE_VERSION " " PACKAGE "/" VERSION "\r\n" );
					if( req->response_headers )
					{
						/* Send custom headers */
						g_list_foreach( req->response_headers, send_header, req );
					}
					nprintf( req->socket, "\r\n" );
				}
				nprintf( req->socket, "%s", req->response->str );
				req->disconnect=TRUE;
				break;
		}
	}

	pthread_cleanup_pop(1);
	return NULL;
}

void request_relocate( struct Request *request, gchar *path )
{
	if( !path || !request ) return;
	if( !strncmp( path, "http://", 7) )
	{
		request->response_headers=g_list_prepend( request->response_headers, g_strdup_printf( "Location: %s\r\n", path ) );
	}
	else /* If we supply only the path and not host or protocol we add the servers address and http */
	{
		if( request->server->port!=80 )
		{
			request->response_headers=g_list_prepend( request->response_headers, g_strdup_printf( "Location: http://%s:%d%s", request->server_hostname, request->server->port, path ) );
		}
		else
		{
			request->response_headers=g_list_prepend( request->response_headers, g_strdup_printf( "Location: http://%s%s", request->server_hostname, path ) );
		}
	}
	g_string_sprintf( request->response,
"<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n"
"<HTML><HEAD>\n"
"<TITLE>301 Moved Permanently</TITLE>\n"
"</HEAD><BODY>\n"
"<H1>Moved Permanently</H1>\n"
"The document has moved <A HREF=\"%s\">here</A>.<P>\n"
"</BODY></HTML>\n", path );
	request->ret_code=301;
}

void request_querydecode( GHashTable *hashtable, gchar *query )
{
	gchar **vptr=NULL;
	gint count;
	gchar *cmdptr;
	gint length, length2;
	GString *gs;
	vptr=g_strsplit( (gchar *)query, (gchar *)"&", 4096 );
	for( count=0 ; vptr[count] ; count++ );
	if( count )
	{
		for( count=0; vptr[count]; count++ )
		{
			length=0;
			if( ( cmdptr=strchr( vptr[count], '=' ) ) )
			{
				cmdptr[0]='\0';
				cmdptr++;
				plustospace( cmdptr );
				length=unescape_url( cmdptr );
			}
			plustospace( vptr[count] );
			length2=unescape_url( vptr[count] );
			if( !length ) length=length2;
			g_strdown( vptr[count] );
			gs=g_new0( GString, 1 );
			if( cmdptr )
			{
				/* To be able to have NULLs here we need to store the length too
				 * GString has a len field. Don't use GString functions on it! */
				gs->str=g_new0( gchar, length+1 );
				memcpy( gs->str, cmdptr, length );
				gs->len=length;
			}
			g_hash_table_insert( hashtable, g_strdup( vptr[count] ), gs );
		}
	}
	g_strfreev( vptr );
	vptr=NULL;
}
